import logging
import contextlib

import paramiko
from paramiko.common import (  # noqa pylint: disable=unused-import
    DEFAULT_WINDOW_SIZE,
    DEFAULT_MAX_PACKET_SIZE,
    MAX_WINDOW_SIZE,
)

from typing import Union

from . import _ssh

logger = logging.getLogger(__name__)

MAX_PACKET_SIZE = 10 * (1024 * 1024)  # this one has no paramiko default we can use


@contextlib.contextmanager
def _connect(
    address: str,
    port: Union[int, str],
    connection_timeout: int,
    transfer_timeout: int,
    window_size: int,
    max_packet_size: int,
    username: Union[str, None] = None,
    password: Union[str, None] = None,
    private_key: Union[str, None] = None,
    passphrase: Union[str, None] = None,
    allow_agent: bool = False,
    missing_host_key_policy: paramiko.MissingHostKeyPolicy = paramiko.AutoAddPolicy(),
):
    logger.debug(
        f"creating ssh connection to {address}:{port} with connection timeout {connection_timeout}, "
        f"transfer timeout {transfer_timeout}, window_size {window_size}, and max_packet_size {max_packet_size}"
    )
    client = _ssh._get_client(
        address,
        port,
        username,
        password,
        private_key,
        passphrase,
        allow_agent,
        connection_timeout,
        missing_host_key_policy,
    )
    transport = client.get_transport()
    transport.default_max_packet_size = max_packet_size
    transport.default_window_size = window_size
    sftp_client = client.open_sftp()
    sftp_client.get_channel().settimeout(transfer_timeout)

    try:
        yield sftp_client
    finally:
        sftp_client.close()
        client.close()
