import time
import logging

from contextlib import contextmanager
from typing import List, Dict, Any, Optional

from ssh import ssh_client, exceptions as base_exceptions

from . import exceptions

logger = logging.getLogger(__name__)


@contextmanager
def retry_handler(
    ssh_server,
    *,
    prompt_timeout: int,
    connection_timeout: int,
    banner_timeout: int,
    retries: Optional[int] = None,
    backoff_delay: Optional[int] = 5,
):
    # This wrapper was written to retry connections when the following connection error happens:
    #
    # ssh.exceptions.ConnectionFailed: Connection to [blah] failed.
    # Error reading SSH protocol banner [Errno 104] Connection reset by peer
    retries = retries or 1  # If retries is supplied as 0 or None it should be set to 1.
    if backoff_delay is None:
        backoff_delay = 5
    if retries > 1:
        logger.debug(f"up to {retries} connection attempts will be made...")
    try:
        for attempt in range(1, (retries + 1)):
            try:
                with ssh_client.initialize(
                    **ssh_server,
                    shell_timeout=prompt_timeout,
                    timeout=connection_timeout,
                    banner_timeout=banner_timeout,
                )() as open_connection:
                    yield open_connection
                    break
            except base_exceptions.ConnectionFailed as e:
                if "Connection reset by peer" in str(e):
                    if attempt < retries:
                        logger.error(str(e))
                        logger.info(f"connection attempt {attempt} reset by peer, retrying after backoff delay...")
                        time.sleep(backoff_delay * attempt)
                        continue
                    else:
                        logger.error("final connection attempt has failed!")
                        raise e
                raise e
    finally:
        # No teardown needs to happen here, the initialize() context manager does it all for us.
        pass


def check_server_details(ssh_server):
    missing_mandatory_keys = [key for key in {"address", "port", "username", "base_prompt"} if key not in ssh_server]
    if missing_mandatory_keys:
        raise Exception(f"SSH server details must include the following missing values: {missing_mandatory_keys}.")

    if not any(["password" in ssh_server, "private_key" in ssh_server]):
        raise Exception(
            "SSH server details must include a password or private key for authentication, but neither was provided."
        )


def run(
    ssh_server,
    command: str,
    prompt_timeout: int,
    connection_timeout: int,
    banner_timeout: int,
    retries: Optional[int] = None,
    backoff_delay: Optional[int] = 5,
) -> str:
    check_server_details(ssh_server)

    with retry_handler(
        ssh_server,
        prompt_timeout=prompt_timeout,
        connection_timeout=connection_timeout,
        banner_timeout=banner_timeout,
        retries=retries,
        backoff_delay=backoff_delay,
    ) as open_connection:
        return ssh_client.run(open_connection, command=command).output


def run_list(
    ssh_server,
    commands: List[Dict[str, Any]],
    prompt_timeout: int,
    connection_timeout: int,
    banner_timeout: int,
    retries: Optional[int] = None,
    backoff_delay: Optional[int] = 5,
) -> List[str]:

    check_server_details(ssh_server)

    for index, command in enumerate(commands, 1):
        questions = command.get("questions")
        if questions:
            question_prompts = [question["prompt"] for question in questions]
            repeated_prompts = set([f"{prompt}" for prompt in question_prompts if question_prompts.count(prompt) > 1])
            if repeated_prompts:
                raise ValueError(
                    "Command {index} has multiple questions with the following prompts: {repeated_prompts}".format(
                        index=index, repeated_prompts=", ".join(repeated_prompts)
                    )
                )

    def _process_command(open_connection, command, questions, success):

        # Simple command that returns either the base or a success prompt
        if not questions:
            return ssh_client.run(open_connection, command=command, prompt=success or ssh_server["base_prompt"]).output

        # Otherwise it's a command with questions
        named_group_template = "(?P<{name}>{expression})"
        named_groups = []

        # Each question prompt in the expression is named after its index in the list
        for index, prompt_dict in enumerate(questions):
            named_groups.append(named_group_template.format(name=f"question_{index}", expression=prompt_dict["prompt"]))

        # If success prompt was supplied, add it to the named group expression
        if success:
            named_groups.append(named_group_template.format(name="success", expression=success))

        # Now add a named group for the base prompt
        named_groups.append(named_group_template.format(name="base", expression=ssh_server["base_prompt"]))
        named_group_expression = "({groups})".format(groups="|".join(named_groups))

        command_output = ssh_client.run(open_connection, command=command, prompt=named_group_expression, use_regex=True)

        match_groups = command_output.match_info.groups

        # If the base or success prompt were matched, the command has completed
        if any([all([success, match_groups.get("success")]), match_groups.get("base")]):
            logger.info("response to the command was either the success or base prompt")
            return command_output.output

        # Otherwise we may need to answer some questions for the command to complete
        def _answer_question(match_groups):
            """
            Each time an answer to a question is run, the output could be:

            1) The base prompt OR a success prompt (return)
            2) Another question (continue)

            The same named_group_expression compiled earlier is used to make the decision.

            Because the questions list in the YAML doesn't name the capturing groups, the
            list index of each question is used at the end of its named group.

            The index is retrieved from the matching group name & used to get the answer
            from the original questions list.

            Indices in log & exception messages start at 1 for readability.

            If multiple question prompts are found in a response, the command fails.
            """
            for _ in range(0, len(questions)):
                question_name = next(iter((name for name in match_groups.keys() if match_groups.get(name))))
                question_index = int(question_name.split("_")[1])
                logger.debug(f"response: {repr(match_groups.get(question_name))}")
                logger.debug(f"response matches the prompt for question {question_index + 1}")

                # Send the answer to the current question
                logger.info(f"answering question {question_index + 1}")
                output = ssh_client.run(
                    open_connection,
                    command=questions[question_index]["answer"],
                    prompt=named_group_expression,
                    use_regex=True,
                )
                match_groups = output.match_info.groups

                # The command is complete if either the success or base prompt is found
                if any([match_groups.get(name) for name in {"success", "base"}]):
                    logger.debug(f"response to question {question_index + 1} was either the success or base prompt")
                    logger.debug(f"returning output: {repr(output.output)}")
                    return output.output
                else:
                    logger.info(f"response to question {question_index + 1} was another question")
            else:
                raise exceptions.QuestionsExhausted(f"Command not complete after {len(questions)} were answered.")

        return _answer_question(match_groups)

    with retry_handler(
        ssh_server,
        prompt_timeout=prompt_timeout,
        connection_timeout=connection_timeout,
        banner_timeout=banner_timeout,
        retries=retries,
        backoff_delay=backoff_delay,
    ) as open_connection:
        return [
            _process_command(open_connection, command["command"], command.get("questions"), command.get("success"))
            for command in commands
        ]
