r"""
Client

This module contains actions for running commands on an SSH device, & uploading files over SCP.

!!! danger "Before you Start"

    See [Running Commands](../running_commands) before using actions in this plugin.

"""
import os
import uuid
import logging
import pathlib

from contextlib import contextmanager
from typing import List, Dict, Any, Optional

import paramiko
from scp import SCPClient

from ssh_plugin_driver.client import run, run_list

from ppa import ui

from .records import SSHServer, SCPServer


logger = logging.getLogger(__name__)

WORKSPACE = "/workspace/share"


@contextmanager
def _temp_file(path, contents):
    try:
        logger.debug(f"creating temporary file at {path}...")
        with open(path, "w+") as temp_file:
            temp_file.write(contents)
        yield
    finally:
        logger.debug(f"removing up temporary file at {path}...")
        os.remove(path)


def run_command(
    ssh_server: SSHServer, command: str, prompt_timeout: Optional[int] = 5, connection_timeout: Optional[int] = 10
) -> str:
    """Run a non-interactive command.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __ssh_server:__ an [SSHPasswordAuth](../inputs_outputs/#sshpasswordauth) or [SSHKeyAuth](../inputs_outputs/#sshkeyauth) dictionary

    - __command:__ the command to run

    - __prompt_timeout:__ the number of seconds to wait for the command to complete (defaults to _5_)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to _10_)

??? success "Output"

    A string containing any command output.

??? example

    ```YAML linenums="1"
    - ssh.client.run_command:
        command: ls --color=never -lsah
      load:
        ssh_server: ssh_server
      save: directory_listing
    ```
"""
    # The banner_timeout is supplied as connection_timeout below, as the value is suitable for both.
    return run(ssh_server, command, prompt_timeout, connection_timeout, connection_timeout)


def run_commands(
    ssh_server: SSHServer,
    commands: List[Dict[str, Any]],
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> List[str]:
    r"""Run multiple commands in the same SSH session.

Use [questions](../inputs_outputs/#commandquestion) to handle any interactive prompts the command may trigger.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __ssh_server:__ an [SSHPasswordAuth](../inputs_outputs/#sshpasswordauth) or [SSHKeyAuth](../inputs_outputs/#sshkeyauth) dictionary

    - __commands:__ a list of [Command](../inputs_outputs/#command) dictionaries

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to _5_)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to _10_)

??? success "Output"

    A list of strings containing the sequential output from each command.

??? example "Example 1 - Non-Interactive Commands"

    ```YAML linenums="1"
    - ssh.client.run_commands:
        commands:
          - command: date
          - command: uptime
          - command: lsb_release -c
      load:
        ssh_server: ssh_server
      save: server_check
    ```

??? example "Example 2 - Password Prompt"

    {% raw %}
    ```YAML linenums="1" hl_lines="4 5 6"
    - ssh.client.run_commands:
        commands:
          - command: sudo ls -lsah --color=never
            questions:
              - prompt: "\\[sudo\\] password for .*"
                answer: "{{ ssh_server.password }}"
      load:
        ssh_server: ssh_server
      save: directory_listing
    ```
    {% endraw %}

"""
    # The banner_timeout is supplied as connection_timeout below, as the value is suitable for both.
    return run_list(ssh_server, commands, prompt_timeout, connection_timeout, connection_timeout)


# Why does this action exist in the SSH plugin?
#
# Right now plugins can't access files saved by other plugins, as they operate in separate chrooted jails. This means
# transferring *actual* files from the task container is useless, as the file would have been generated & saved
# inside a different jail.
#
# This architectural limitation means it's not worth making SCP or SFTP plugins right now - we wouldn't have any
# meaningful actions in them. For now, supplying the file contents as text -> creating a temp file -> uploading it
# via SCP is a useful workaround, & it doesn't warrant a new plugin.
#
# This is also why I haven't put this in the SSH plugin driver SDK; it shouldn't exist for much longer.
def scp_upload_from_text(scp_server: SCPServer, file_contents: str, destination: str) -> None:
    r"""Upload some text as a file to an SCP server.

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __scp_server:__ an [SCPServer](../inputs_outputs/#scpserver) dictionary

    - __file_contents:__ the contents of the text file

    - __destination:__ the destination path on the SCP server (_must include the file name_)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - ssh.client.scp_upload_from_text:
        file_contents: >
          This will be uploaded as a text file.
        destination: /home/ppa/text-file.txt
      load:
        ssh_server: ssh_server
    ```
"""

    local_file_path = pathlib.Path(WORKSPACE) / f"{str(uuid.uuid4())}.txt"
    logger.info(f"the data will be temporarily stored at: {local_file_path.as_posix()}")

    try:
        scp_client = paramiko.SSHClient()
        scp_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        scp_client.connect(
            scp_server["address"],
            port=scp_server["port"],
            username=scp_server["username"],
            password=scp_server["password"],
            allow_agent=False,
            look_for_keys=False,
        )

        with SCPClient(scp_client.get_transport()) as scp_connection:
            with _temp_file(local_file_path, file_contents):
                logger.info(f"uploading command data via scp to {scp_server.get('address')}...")
                scp_connection.put(local_file_path, destination)
                logger.info(f"scp upload to {scp_server.get('address')} complete")

    except paramiko.ssh_exception.AuthenticationException as e:
        logger.exception(e)
        ui.output_error(f"Failed to authenticate to SCP server {scp_server.get('address')}:{scp_server.get('port')}.")
        raise e

    except paramiko.ssh_exception.NoValidConnectionsError as e:
        logger.exception(e)
        ui.output_error(f"Could not connect to SCP server {scp_server.get('address')}:{scp_server.get('port')}.")
        ui.output_error("Please view the logs for more information.")
        raise e
